/*
 * Minecraft Forge
 * Copyright (c) 2016-2019.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation version 2.1
 * of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package net.minecraftforge.client;

import javax.annotation.Nonnull;
import java.nio.ByteBuffer;
import java.util.function.Predicate;

import org.lwjgl.opengl.GL11;

import com.mojang.blaze3d.platform.GLX;
import com.mojang.blaze3d.platform.GlStateManager;

import net.minecraft.client.Minecraft;
import net.minecraft.client.world.ClientWorld;
import net.minecraft.client.renderer.BufferBuilder;
import net.minecraft.client.renderer.GLAllocation;
import net.minecraft.client.renderer.Tessellator;
import net.minecraft.client.renderer.texture.DynamicTexture;
import net.minecraft.client.renderer.vertex.DefaultVertexFormats;
import net.minecraft.client.renderer.vertex.VertexBuffer;
import net.minecraft.client.renderer.vertex.VertexFormat;
import net.minecraft.client.settings.CloudOption;
import net.minecraft.entity.Entity;
import net.minecraft.resources.IReloadableResourceManager;
import net.minecraft.resources.IResourceManager;
import net.minecraft.util.ResourceLocation;
import net.minecraft.util.math.MathHelper;
import net.minecraft.util.math.Vec3d;
import net.minecraftforge.common.ForgeConfig;
import net.minecraftforge.resource.IResourceType;
import net.minecraftforge.resource.ISelectiveResourceReloadListener;
import net.minecraftforge.resource.VanillaResourceType;

public class CloudRenderer implements ISelectiveResourceReloadListener
{
    // Shared constants.
    private static final float PX_SIZE = 1 / 256F;

    // Building constants.
    private static final VertexFormat FORMAT = DefaultVertexFormats.field_181709_i;
    private static final int TOP_SECTIONS = 12;    // Number of slices a top face will span.
    private static final int HEIGHT = 4;
    private static final float INSET = 0.001F;
    private static final float ALPHA = 0.8F;

    // Debug
    private static final boolean WIREFRAME = false;

    // Instance fields
    private final Minecraft mc = Minecraft.func_71410_x();
    private final ResourceLocation texture = new ResourceLocation("textures/environment/clouds.png");

    private int displayList = -1;
    private VertexBuffer vbo;
    private CloudOption cloudMode = CloudOption.OFF;
    private int renderDistance = -1;

    private DynamicTexture COLOR_TEX = null;

    private int texW;
    private int texH;

    public CloudRenderer()
    {
        // Resource manager should always be reloadable.
        ((IReloadableResourceManager) mc.func_195551_G()).func_219534_a(this);
    }

    private int getScale()
    {
        return cloudMode == CloudOption.FANCY ? 12 : 8;
    }

    private float ceilToScale(float value)
    {
        float scale = getScale();
        return MathHelper.func_76123_f(value / scale) * scale;
    }

    private void vertices(BufferBuilder buffer)
    {
        boolean fancy = cloudMode == CloudOption.FANCY;    // Defines whether to hide all but the bottom.

        float scale = getScale();
        float CULL_DIST = 2 * scale;

        float bCol = fancy ? 0.7F : 1F;

        float sectEnd = ceilToScale((renderDistance * 2) * 16);
        float sectStart = -sectEnd;

        float sectStep = ceilToScale(sectEnd * 2 / TOP_SECTIONS);
        float sectPx = PX_SIZE / scale;

        buffer.func_181668_a(GL11.GL_QUADS, FORMAT);

        float sectX0 = sectStart;
        float sectX1 = sectX0;

        while (sectX1 < sectEnd)
        {
            sectX1 += sectStep;

            if (sectX1 > sectEnd)
                sectX1 = sectEnd;

            float sectZ0 = sectStart;
            float sectZ1 = sectZ0;

            while (sectZ1 < sectEnd)
            {
                sectZ1 += sectStep;

                if (sectZ1 > sectEnd)
                    sectZ1 = sectEnd;

                float u0 = sectX0 * sectPx;
                float u1 = sectX1 * sectPx;
                float v0 = sectZ0 * sectPx;
                float v1 = sectZ1 * sectPx;

                // Bottom
                buffer.func_181662_b(sectX0, 0, sectZ0).func_187315_a(u0, v0).func_181666_a(bCol, bCol, bCol, ALPHA).func_181675_d();
                buffer.func_181662_b(sectX1, 0, sectZ0).func_187315_a(u1, v0).func_181666_a(bCol, bCol, bCol, ALPHA).func_181675_d();
                buffer.func_181662_b(sectX1, 0, sectZ1).func_187315_a(u1, v1).func_181666_a(bCol, bCol, bCol, ALPHA).func_181675_d();
                buffer.func_181662_b(sectX0, 0, sectZ1).func_187315_a(u0, v1).func_181666_a(bCol, bCol, bCol, ALPHA).func_181675_d();

                if (fancy)
                {
                    // Top
                    buffer.func_181662_b(sectX0, HEIGHT, sectZ0).func_187315_a(u0, v0).func_181666_a(1, 1, 1, ALPHA).func_181675_d();
                    buffer.func_181662_b(sectX0, HEIGHT, sectZ1).func_187315_a(u0, v1).func_181666_a(1, 1, 1, ALPHA).func_181675_d();
                    buffer.func_181662_b(sectX1, HEIGHT, sectZ1).func_187315_a(u1, v1).func_181666_a(1, 1, 1, ALPHA).func_181675_d();
                    buffer.func_181662_b(sectX1, HEIGHT, sectZ0).func_187315_a(u1, v0).func_181666_a(1, 1, 1, ALPHA).func_181675_d();

                    float slice;
                    float sliceCoord0;
                    float sliceCoord1;

                    for (slice = sectX0; slice < sectX1;)
                    {
                        sliceCoord0 = slice * sectPx;
                        sliceCoord1 = sliceCoord0 + PX_SIZE;

                        // X sides
                        if (slice > -CULL_DIST)
                        {
                            slice += INSET;
                            buffer.func_181662_b(slice, 0,      sectZ1).func_187315_a(sliceCoord0, v1).func_181666_a(0.9F, 0.9F, 0.9F, ALPHA).func_181675_d();
                            buffer.func_181662_b(slice, HEIGHT, sectZ1).func_187315_a(sliceCoord1, v1).func_181666_a(0.9F, 0.9F, 0.9F, ALPHA).func_181675_d();
                            buffer.func_181662_b(slice, HEIGHT, sectZ0).func_187315_a(sliceCoord1, v0).func_181666_a(0.9F, 0.9F, 0.9F, ALPHA).func_181675_d();
                            buffer.func_181662_b(slice, 0,      sectZ0).func_187315_a(sliceCoord0, v0).func_181666_a(0.9F, 0.9F, 0.9F, ALPHA).func_181675_d();
                            slice -= INSET;
                        }

                        slice += scale;

                        if (slice <= CULL_DIST)
                        {
                            slice -= INSET;
                            buffer.func_181662_b(slice, 0,      sectZ0).func_187315_a(sliceCoord0, v0).func_181666_a(0.9F, 0.9F, 0.9F, ALPHA).func_181675_d();
                            buffer.func_181662_b(slice, HEIGHT, sectZ0).func_187315_a(sliceCoord1, v0).func_181666_a(0.9F, 0.9F, 0.9F, ALPHA).func_181675_d();
                            buffer.func_181662_b(slice, HEIGHT, sectZ1).func_187315_a(sliceCoord1, v1).func_181666_a(0.9F, 0.9F, 0.9F, ALPHA).func_181675_d();
                            buffer.func_181662_b(slice, 0,      sectZ1).func_187315_a(sliceCoord0, v1).func_181666_a(0.9F, 0.9F, 0.9F, ALPHA).func_181675_d();
                            slice += INSET;
                        }
                    }

                    for (slice = sectZ0; slice < sectZ1;)
                    {
                        sliceCoord0 = slice * sectPx;
                        sliceCoord1 = sliceCoord0 + PX_SIZE;

                        // Z sides
                        if (slice > -CULL_DIST)
                        {
                            slice += INSET;
                            buffer.func_181662_b(sectX0, 0,      slice).func_187315_a(u0, sliceCoord0).func_181666_a(0.8F, 0.8F, 0.8F, ALPHA).func_181675_d();
                            buffer.func_181662_b(sectX0, HEIGHT, slice).func_187315_a(u0, sliceCoord1).func_181666_a(0.8F, 0.8F, 0.8F, ALPHA).func_181675_d();
                            buffer.func_181662_b(sectX1, HEIGHT, slice).func_187315_a(u1, sliceCoord1).func_181666_a(0.8F, 0.8F, 0.8F, ALPHA).func_181675_d();
                            buffer.func_181662_b(sectX1, 0,      slice).func_187315_a(u1, sliceCoord0).func_181666_a(0.8F, 0.8F, 0.8F, ALPHA).func_181675_d();
                            slice -= INSET;
                        }

                        slice += scale;

                        if (slice <= CULL_DIST)
                        {
                            slice -= INSET;
                            buffer.func_181662_b(sectX1, 0,      slice).func_187315_a(u1, sliceCoord0).func_181666_a(0.8F, 0.8F, 0.8F, ALPHA).func_181675_d();
                            buffer.func_181662_b(sectX1, HEIGHT, slice).func_187315_a(u1, sliceCoord1).func_181666_a(0.8F, 0.8F, 0.8F, ALPHA).func_181675_d();
                            buffer.func_181662_b(sectX0, HEIGHT, slice).func_187315_a(u0, sliceCoord1).func_181666_a(0.8F, 0.8F, 0.8F, ALPHA).func_181675_d();
                            buffer.func_181662_b(sectX0, 0,      slice).func_187315_a(u0, sliceCoord0).func_181666_a(0.8F, 0.8F, 0.8F, ALPHA).func_181675_d();
                            slice += INSET;
                        }
                    }
                }

                sectZ0 = sectZ1;
            }

            sectX0 = sectX1;
        }
    }

    private void dispose()
    {
        if (vbo != null)
        {
            vbo.func_177362_c();
            vbo = null;
        }
        if (displayList >= 0)
        {
            GLAllocation.func_74523_b(displayList);
            displayList = -1;
        }
    }

    private void build()
    {
        Tessellator tess = Tessellator.func_178181_a();
        BufferBuilder buffer = tess.func_178180_c();

        if (GLX.useVbo())
            vbo = new VertexBuffer(FORMAT);
        else
            GlStateManager.newList(displayList = GLAllocation.func_74526_a(1), GL11.GL_COMPILE);

        vertices(buffer);

        if (GLX.useVbo())
        {
            buffer.func_178977_d();
            buffer.func_178965_a();
            vbo.func_181722_a(buffer.func_178966_f());
        }
        else
        {
            tess.func_78381_a();
            GlStateManager.endList();
        }
    }

    private int fullCoord(double coord, int scale)
    {   // Corrects misalignment of UV offset when on negative coords.
        return ((int) coord / scale) - (coord < 0 ? 1 : 0);
    }

    private boolean isBuilt()
    {
        return GLX.useVbo() ? vbo != null : displayList >= 0;
    }

    public void checkSettings()
    {
        boolean newEnabled = ForgeConfig.CLIENT.forgeCloudsEnabled.get()
                && mc.field_71474_y.func_216842_e() != CloudOption.OFF
                && mc.field_71441_e != null
                && mc.field_71441_e.field_73011_w.func_76569_d();

        if (isBuilt()
                    && (!newEnabled
                    || mc.field_71474_y.func_216842_e() != cloudMode
                    || mc.field_71474_y.field_151451_c != renderDistance))
        {
            dispose();
        }

        cloudMode = mc.field_71474_y.func_216842_e();
        renderDistance = mc.field_71474_y.field_151451_c;

        if (newEnabled && !isBuilt())
        {
            build();
        }
    }

    public boolean render(int cloudTicks, float partialTicks)
    {
        if (!isBuilt())
            return false;

        Entity entity = mc.func_175606_aa();

        double totalOffset = cloudTicks + partialTicks;

        double x = entity.field_70169_q + (entity.field_70165_t - entity.field_70169_q) * partialTicks
                + totalOffset * 0.03;
        double y = mc.field_71441_e.field_73011_w.func_76571_f()
                - (entity.field_70137_T + (entity.field_70163_u - entity.field_70137_T) * partialTicks)
                + 0.33;
        double z = entity.field_70166_s + (entity.field_70161_v - entity.field_70166_s) * partialTicks;

        int scale = getScale();

        if (cloudMode == CloudOption.FANCY)
            z += 0.33 * scale;

        // Integer UVs to translate the texture matrix by.
        int offU = fullCoord(x, scale);
        int offV = fullCoord(z, scale);

        GlStateManager.pushMatrix();

        // Translate by the remainder after the UV offset.
        GlStateManager.translated((offU * scale) - x, y, (offV * scale) - z);

        // Modulo to prevent texture samples becoming inaccurate at extreme offsets.
        offU = offU % texW;
        offV = offV % texH;

        // Translate the texture.
        GlStateManager.matrixMode(GL11.GL_TEXTURE);
        GlStateManager.translatef(offU * PX_SIZE, offV * PX_SIZE, 0);
        GlStateManager.matrixMode(GL11.GL_MODELVIEW);

        GlStateManager.disableCull();

        GlStateManager.enableBlend();
        GlStateManager.blendFuncSeparate(
                GlStateManager.SourceFactor.SRC_ALPHA, GlStateManager.DestFactor.ONE_MINUS_SRC_ALPHA,
                GlStateManager.SourceFactor.ONE, GlStateManager.DestFactor.ZERO);

        // Color multiplier.
        Vec3d color = mc.field_71441_e.func_72824_f(partialTicks);
        float r = (float) color.field_72450_a;
        float g = (float) color.field_72448_b;
        float b = (float) color.field_72449_c;

        if (COLOR_TEX == null)
            COLOR_TEX = new DynamicTexture(1, 1, false);

        // Apply a color multiplier through a texture upload if shaders aren't supported.
        COLOR_TEX.func_195414_e().func_195700_a(0, 0, 255 << 24
                | ((int) (r * 255)) << 16
                | ((int) (g * 255)) << 8
                | (int) (b * 255));
        COLOR_TEX.func_110564_a();

        GlStateManager.activeTexture(GLX.GL_TEXTURE1);
        GlStateManager.bindTexture(COLOR_TEX.func_110552_b());
        GlStateManager.enableTexture();

        // Bind the clouds texture last so the shader's sampler2D is correct.
        GlStateManager.activeTexture(GLX.GL_TEXTURE0);
        mc.field_71446_o.func_110577_a(texture);

        ByteBuffer buffer = Tessellator.func_178181_a().func_178180_c().func_178966_f();

        // Set up pointers for the display list/VBO.
        if (GLX.useVbo())
        {
            vbo.func_177359_a();

            int stride = FORMAT.func_177338_f();
            GlStateManager.vertexPointer(3, GL11.GL_FLOAT, stride, 0);
            GlStateManager.enableClientState(GL11.GL_VERTEX_ARRAY);
            GlStateManager.texCoordPointer(2, GL11.GL_FLOAT, stride, 12);
            GlStateManager.enableClientState(GL11.GL_TEXTURE_COORD_ARRAY);
            GlStateManager.colorPointer(4, GL11.GL_UNSIGNED_BYTE, stride, 20);
            GlStateManager.enableClientState(GL11.GL_COLOR_ARRAY);
        }
        else
        {
            buffer.limit(FORMAT.func_177338_f());
            for (int i = 0; i < FORMAT.func_177345_h(); i++)
                FORMAT.func_177343_g().get(i).func_177375_c().preDraw(FORMAT, i, FORMAT.func_177338_f(), buffer);
            buffer.position(0);
        }

        // Depth pass to prevent insides rendering from the outside.
        GlStateManager.colorMask(false, false, false, false);
        if (GLX.useVbo())
            vbo.func_177358_a(GL11.GL_QUADS);
        else
            GlStateManager.callList(displayList);

        // Full render.
        GlStateManager.colorMask(true, true, true, true);

        // Wireframe for debug.
        if (WIREFRAME)
        {
            GlStateManager.polygonMode(GL11.GL_FRONT_AND_BACK, GL11.GL_LINE);
            GlStateManager.lineWidth(2.0F);
            GlStateManager.disableTexture();
            GlStateManager.depthMask(false);
            GlStateManager.disableFog();
            if (GLX.useVbo())
                vbo.func_177358_a(GL11.GL_QUADS);
            else
                GlStateManager.callList(displayList);
            GlStateManager.polygonMode(GL11.GL_FRONT_AND_BACK, GL11.GL_FILL);
            GlStateManager.depthMask(true);
            GlStateManager.enableTexture();
            GlStateManager.enableFog();
        }

        if (GLX.useVbo())
        {
            vbo.func_177358_a(GL11.GL_QUADS);
            vbo.func_177361_b(); // Unbind buffer and disable pointers.
        }
        else
        {
            GlStateManager.callList(displayList);
        }

        buffer.limit(0);
        for (int i = 0; i < FORMAT.func_177345_h(); i++)
            FORMAT.func_177343_g().get(i).func_177375_c().postDraw(FORMAT, i, FORMAT.func_177338_f(), buffer);
        buffer.position(0);

        // Disable our coloring.
        GlStateManager.activeTexture(GLX.GL_TEXTURE1);
        GlStateManager.disableTexture();
        GlStateManager.activeTexture(GLX.GL_TEXTURE0);

        // Reset texture matrix.
        GlStateManager.matrixMode(GL11.GL_TEXTURE);
        GlStateManager.loadIdentity();
        GlStateManager.matrixMode(GL11.GL_MODELVIEW);

        GlStateManager.disableBlend();
        GlStateManager.enableCull();

        GlStateManager.popMatrix();

        return true;
    }

    private void reloadTextures()
    {
        if (mc.field_71446_o != null)
        {
            mc.field_71446_o.func_110577_a(texture);
            texW = GlStateManager.getTexLevelParameter(GL11.GL_TEXTURE_2D, 0, GL11.GL_TEXTURE_WIDTH);
            texH = GlStateManager.getTexLevelParameter(GL11.GL_TEXTURE_2D, 0, GL11.GL_TEXTURE_HEIGHT);
        }
    }

    @Override
    public void onResourceManagerReload(@Nonnull IResourceManager resourceManager, @Nonnull Predicate<IResourceType> resourcePredicate)
    {
        if (resourcePredicate.test(VanillaResourceType.TEXTURES))
        {
            reloadTextures();
        }
    }

    private static CloudRenderer cloudRenderer;
    private static CloudRenderer getCloudRenderer()
    {
        if (cloudRenderer == null)
            cloudRenderer = new CloudRenderer();
        return cloudRenderer;
    }

    public static void updateCloudSettings()
    {
        getCloudRenderer().checkSettings();
    }

    public static boolean renderClouds(int cloudTicks, float partialTicks, ClientWorld world, Minecraft client)
    {
        IRenderHandler renderer = world.field_73011_w.getCloudRenderer();
        if (renderer != null)
        {
            renderer.render(partialTicks, world, client);
            return true;
        }
        return getCloudRenderer().render(cloudTicks, partialTicks);
    }

}
